/**
 * Brightness control daemon
 * For Thinkpad laptops
 * Copyright (c) 2006, Phillip Berndt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Compile using:
 *  gcc -lX11 -lXss bright.c
 *
 * vim:ft=c:fileencoding=iso-8859-1
 */
#define RELEASE "0.3_pre1"

#include <unistd.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/scrnsaver.h>
#include <dirent.h>
Display *display;

/**
 * Overall program settings
 */
int verbose     = 0;
int daemonize   = 0;
int waitSeconds = 3;
int speed       = 3;
int darkBright  = 0;
int force       = 0;
int maxBright   = 0;

char b_class[21];
char actualBrightnessFile[255];
char brightnessFile[255];

/*
 * Write an error to stderr and exit
 */
void error(char *str)
{
	if(daemonize == 1)
	{
		return;
	}

	fputs(str, stderr);
	fputs("\n", stderr);

	/* Close display */
	if(display)
	{
		XCloseDisplay(display);
	}

	/* Quit */
	exit(1);
}

/*
 * Write an debug message to the console
 */
void info(char *str)
{
	if(verbose == 0 || daemonize == 1)
	{
		return;
	}

	printf("%s\n", str);
}

/*
 * Function to set the brightness of the IBM
 * display
 */
void setBrightness(int level)
{
	char output[255];
	FILE *brightness = fopen(brightnessFile, "w");
	if(!brightness)
	{
		error("Failed to open /sys/class/backlight/*/brightness for writing.\n "
			"- Do you have the (IBM) ACPI extensions for the class supplied "
			"using -c in your kernel?\n"
			"- Do you have permissions to write to that file?");
	}
	sprintf(output, "%i\n", level);
	fputs(output, brightness);
	fclose(brightness);
}

/*
 * Function to get the brightness of the pc
 */
int getBrightness()
{
	char line[255];
	FILE *brightness = fopen(actualBrightnessFile, "r");
	if(!brightness)
	{
		error("Failed to open /sys/class/backlight/*/actual_brightness. Do you have the "
			"ACPI extensions for the class supplied using -c in your kernel?");
	}
	fgets(line, 255, brightness);
	int retVal = atoi(line);
	fclose(brightness);

	return retVal;
}

/*
 * Check whether the PC is on AC
 */
int isOnAC()
{
	char line[255];
	FILE *ac = fopen("/proc/acpi/ac_adapter/AC/state", "r");
	if(!ac)
	{
		/* Ignore this */
		return 0;
	}
	fgets(line, 255, ac);
	fclose(ac);

	return strstr(line, "on-line") != NULL;
}

/*
 * Signal handler
 */
void signalhandler(int signal)
{
	error("Received signal. Exiting...");
}

/*
 * Get IDLE time
 * Taken from GAJIM source, src/common/idle.c
 */
int getIdleTime()
{
	static XScreenSaverInfo *mit_info = NULL;
	int idle_time, event_base, error_base;
	if (XScreenSaverQueryExtension(display, &event_base, &error_base))
	{
		if (mit_info == NULL)
		{
			mit_info = XScreenSaverAllocInfo();
		}
		XScreenSaverQueryInfo(display, RootWindow(display, 0), mit_info);

		if(mit_info->state == 3) /* ScreenSaverDisabled */
		{
			idle_time = 0;
		}
		else
		{
			idle_time = (mit_info->idle) / 1000;
		}
	}
	else
	{
		idle_time = 0;
	}
	return idle_time;
}

/*
 * Print a help message and exit
 */
void printHelp()
{
	printf("brightd %s\nA X11-daemon for iBook-like brightness management\nCopyright (c) 2006-2007, Phillip Berndt\n\nOptions:\n", RELEASE);
	printf(" -v        Be verbose\n");
	printf(" -d        Daemonize\n");
	printf(" -w n      Wait n seconds before reducing brightness (Defaults to 3)\n");
	printf(" -s n      Speed of fade (0-10, defaults to 3, 0 means fast)\n");
	printf(" -b n      The brightness setting for the dark screen (Defaults zu 0)\n");
	printf(" -f        Reduce brightness even if on the highest brightness level or on AC\n");
	printf(" -c n      Set the backlight class to use (defaults to the first subnode of /sys/class/backlight)\n");
	printf("\n");
	exit(0);
}

/*
 * Load default brightness class
 */
void loadDefaultClass()
{
	struct dirent *dirEntry;
	DIR *backlightDir = opendir("/sys/class/backlight");
	if(backlightDir)
	{
		while(1)
		{
			dirEntry = readdir(backlightDir);
			if(!dirEntry)
			{
				break;
			}
			if(dirEntry->d_name[0] == '.')
			{
				continue;
			}
			strcpy(b_class, dirEntry->d_name);
			closedir(backlightDir);
			return;
		}
		closedir(backlightDir);
	}
	strcpy(b_class, "none");
}

/*
 * The main program
 */
int main(int argc, char *argv[])
{
	int  idle              = 0;
	int  i;
	int  dark              = 0;
	int  defaultBrightness = 0;
	int  oldBrightness     = 0;
	int  newBrightness     = 0;
	char option;
	FILE *openTest;

	/* Load settings */
	loadDefaultClass();
	opterr = 0;
	if(argc == 0)
	{
		printHelp();
	}
	while((option = getopt(argc, argv, "vdfw:s:b:c:")) > 0)
	{
		switch(option)
		{
			case 'v':
				verbose = 1;
				break;
			case 'w':
				waitSeconds = atoi(optarg);
				if(waitSeconds <= 0)
				{
					printHelp();
				}
				break;
			case 's':
				speed = atoi(optarg);
				if(speed < 0 || speed > 10)
				{
					printHelp();
				}
				break;
			case 'd':
				daemonize = 1;
				break;
			case 'b':
				darkBright = atoi(optarg);
				if(darkBright < 0 || darkBright > 5)
				{
					printHelp();
				}
				break;
			case 'f':
				force = 1;
				break;
			case 'c':
				if(strlen(optarg) > 20)
				{
					error("The class should not be longer than 20 characters.");
				}
				strcpy(b_class, optarg);
				break;
			default:
				printHelp();
		}
	}

	/* Set files in sysfs to use */
	sprintf(brightnessFile, "/sys/class/backlight/%s/brightness", b_class);
	openTest = fopen(brightnessFile, "r");
	if(!openTest) {
		printf("Failed to open %s\n", brightnessFile);
		error("The specified brightness class was not found. Did you compile the ACPI extensions "
			"for your system into your kernel?");
	}
	fclose(openTest);

	/* Load maximum brightness; abuse some not yet initialised variables to do so ;) */
	sprintf(actualBrightnessFile, "/sys/class/backlight/%s/max_brightness", b_class);
	maxBright = getBrightness();
	if(verbose == 1 && daemonize == 0)
	{
		printf("Maximum brightness is %d\n", maxBright);
	}
	sprintf(actualBrightnessFile, "/sys/class/backlight/%s/actual_brightness", b_class);

	/* Load brightness */
	defaultBrightness = newBrightness = oldBrightness = getBrightness();

	/* Open display */
	display = XOpenDisplay(NULL);
	if(!display)
	{
		error("Failed to open display");
	}

	/* Daemonize */
	if(daemonize == 1)
	{
		daemon(0, 0);
	}

	/* Create signal handler for closing the display */
	signal(SIGINT, signalhandler);
	signal(SIGHUP, signalhandler);

	/* Now loop and wait for what happens ;) */
	while(1)
	{
		/* Do nothing if on AC */
		if(isOnAC() &! force)
		{
			if(dark == 1)
			{
				info("AC plugged in.");
				setBrightness(defaultBrightness);
				dark = 0;
			}

			oldBrightness = getBrightness();
			sleep(2);
			continue;
		}

		/* Check for changes in default brightness */
		newBrightness = getBrightness();
		if(newBrightness - oldBrightness != 0)
		{
			if(dark == 1)
			{
				/* Do nothing in "force" mode */
				if(force == 1)
				{
					setBrightness(darkBright);
					info("Ignoring brightness change in force mode");
					continue;
				}

				/* Change to highest brightness */
				setBrightness(maxBright);
				dark = 0;
				info("Change to highest brightness");
			}
			else
			{
				/* Disallow changing the brightness to a value below the minimum */
				if(newBrightness < darkBright)
				{
					setBrightness(darkBright);
					info("Ignoring brightness change below minimum");
					continue;
				}

				/* Set new default brightness */
				defaultBrightness = newBrightness;
				if(verbose == 1 && daemonize == 0)
				{
					printf("User changed brightness by %d\n", (newBrightness - oldBrightness));
				}
				oldBrightness = newBrightness;
			}
			continue;
		}

		/* Get idle time and change brightness if needed */
		idle = getIdleTime();
		if(idle >= waitSeconds && dark == 0 && (force || newBrightness < maxBright))
		{
			info("Fading to dark setting");
			for(i=newBrightness; i>=darkBright; i--)
			{
				setBrightness(i);
				usleep(speed * 1000);
			}
			dark = 1;
		}
		else if(idle < waitSeconds && dark == 1)
		{
			info("Fading to bright");
			for(i=newBrightness; i<=defaultBrightness; i++)
			{
				setBrightness(i);
				usleep(speed * 1000);
			}
			dark = 0;
		}

		newBrightness = getBrightness();
		oldBrightness = newBrightness;
		usleep(300);
	}

}
